/*
 * Propeller Game Engine (C) 2013 Marco Maccaferri
 *
 * SID/MOS8580 emulator
 * Based on SIDcog by Johannes Ahlebrand
 *
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
 * is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Contributors:
 *     Johannes Ahlebrand - Initial implementation
 *     Marco Maccaferri   - Converted to C and adapted for game engine
 */

#ifndef SID_H_
#define SID_H_

#define TRIANGLE    16
#define SAW         32
#define SQUARE      64
#define NOISE       128

typedef union {
    uint8_t reg[28];
    struct {
        struct {
            uint16_t frequency;
            uint16_t pulseWidth;
            uint8_t  controlRegister;
            uint8_t  attackDecay;
            uint8_t  sustainRelease;
            uint8_t  dummy;
        } ch[3];
        uint8_t Filter1;
        uint8_t Filter2;
        uint8_t Filter3;
        uint8_t Volume;
    };
} SID_REGISTERS;

/**
 * Starts SIDcog in a single cog
 *
 * Returns a pointer to the first SID register in hub memory
 * on success; otherwise returns 0.
 *
 * right - The pin to output the right channel to. 0 = Not used
 * left - The pin to output the left channel to. 0 = Not used
 */
SID_REGISTERS * sid_start(int16_t Right, int16_t Left);

/**
 * Reset all 25 SID registers
 */
void sid_reset_registers();

/**
 * Sets a single SID register to a value
 *
 * reg - The SID register to set
 * val - The value to set the register to
 */
void sid_set_register(int32_t Reg, int32_t Val);

/**
 * Sets the main volume
 *
 * value - A value betwen 0 and 15
 */
void sid_set_volume(int32_t Volumevalue);

/**
 * Plays a tone in a SID channel
 *
 * channel - The SID channel to use. (0 - 2)
 * freq - The 16 bit frequency value use. (0 - 65535)
 *        (The SID can output tone frequencies from 0 - 3.9 kHz)
 * waveform - The waveform combination to use
 *            e.g. sid.play(x, x, SQUARE | SAW, x, x, x, x)
 * attack - The attack value. (0 - 15)
 * decay - The decay value. (0 - 15)
 * sustain - The sustain value. (0 - 15)
 * release - The release value. (0 - 15)
 *
 * When calling this method, the envelope generator enters the
 * "attack - decay - sustain" phase. Don't forget to call
 * "noteOff" before using it so the envelope is in release phase
 */
void sid_play(int32_t Channel, int32_t Freq, int32_t Waveform, int32_t Attack, int32_t Decay, int32_t Sustain, int32_t Release);

/**
 * Plays a tone in a SID channel
 *
 * channel - The SID channel to use. (0 - 2)
 * freq - The 16 bit frequency value use. (0 - 65535)
 *        (The SID can output tone frequencies from 0 - 3.9 kHz)
 *
 * Don't forget to set the envelope values for the channel
 * before using this method.
 *
 * Make sure you have set the waveform for the channel before
 * using this method.
 *
 * When calling this method, the envelope generator enters the
 * "attack - decay - sustain" phase. Don't forget to call
 * "noteOff" before calling this method to set the envelope to
 * release phase.
 */
void sid_noteon(int32_t Channel, int32_t Freq);

/**
 * Sets the envelope generator of a channel to release phase
 *
 * channel - The SID channel to use. (0 - 2)
 */
void sid_noteoff(int32_t Channel);

/**
 * Sets the frequency of a SID channel
 *
 * channel - The SID channel to use. (0 - 2)
 * freq - The 16 bit frequency value use. (0 - 65535)
 *        (The SID can output tone frequencies from 0 - 3.9 kHz)
 */
void sid_setfreq(int32_t Channel, int32_t Freq);

/**
 * Sets the waveform of a SID channel
 *
 * channel - The SID channel to use. (0 - 2)
 * waveform - The waveform combination to use
 *            e.g. sid.play(x, x, SQUARE | SAW, x, x, x, x)
 */
void sid_set_waveform(int32_t Channel, int32_t Waveform);

/**
 * Sets the pulse width of a SID channel
 *
 * channel - The SID channel to use. (0 - 2)
 * pulseWidth - The 12 bit pulse width value to use. (0 - 4095)
 *
 * The pulse width value affects square waves ONLY.
 */
void sid_set_pulse_width(int32_t Channel, int32_t Pulsewidth);

/**
 * Sets the envelope values of a SID channel
 *
 * channel - The SID channel to use. (0 - 2)
 * attack - The attack value. (0 - 15)
 * decay - The decay value. (0 - 15)
 * sustain - The sustain value. (0 - 15)
 * release - The release value. (0 - 15)
 */
void sid_set_adsr(int32_t Channel, int32_t Attack, int32_t Decay, int32_t Sustain, int32_t Release);

/**
 * Sets the resonance value of the filter
 *
 * resonanceValue - The resonance value to use. (0 - 15)
 */
void sid_set_resonance(int32_t Resonancevalue);

/**
 * Sets the cutoff frequency of the filter
 *
 * cutoffValue - The 12 bit cutoff frequency value to use.
 */
void sid_set_cutoff(int32_t Cutoffvalue);

/**
 * Enable/Disable filtering on channels
 *
 * ch1 - Enable/Disable filter on channel 1. (True/False)
 * ch2 - Enable/Disable filter on channel 2. (True/False)
 * ch3 - Enable/Disable filter on channel 3. (True/False)
 */
void sid_set_filter_mask(int32_t Ch1, int32_t Ch2, int32_t Ch3);

/**
 * Enable/Disable filter types
 *
 * lp - Enable/Disable lowpass filter. (True/False)
 * bp - Enable/Disable bandpass filter. (True/False)
 * hp - Enable/Disable highpass filter. (True/False)
 */
void sid_set_filter_type(int32_t Lp, int32_t Bp, int32_t Hp);

/**
 * Enable/Disable ring modulation on channels
 *
 * ch1 - Enable/Disable ring modulation on ch 1. (True/False)
 * ch2 - Enable/Disable ring modulation on ch 2. (True/False)
 * ch3 - Enable/Disable ring modulation on ch 3. (True/False)
 *
 * Channel 3 modulates channel 1
 * Channel 1 modulates channel 2
 * Channel 2 modulates channel 3
 */
void sid_enable_ringmod(int32_t Ch1, int32_t Ch2, int32_t Ch3);

/**
 * Enable/Disable oscillator synchronization on channels
 *
 * ch1 - Enable/Disable synchronization on ch 1. (True/False)
 * ch2 - Enable/Disable synchronization on ch 2. (True/False)
 * ch3 - Enable/Disable synchronization on ch 3. (True/False)
 *
 * Channel 3 synchronizes channel 1
 * Channel 1 synchronizes channel 2
 * Channel 2 synchronizes channel 3
 */
void sidn_enable_synchronization(int32_t Ch1, int32_t Ch2, int32_t Ch3);

#endif /* SID_H_ */
